#!/usr/bin/env node
var express = require('express');
var app = express();
var http = require('http').Server(app);
var io = require('socket.io')(http);
const { v4: uuidv4 } = require('uuid');
const fsExtra = require('fs-extra');
const homedir = require('os').homedir();

var argv = require('minimist')(process.argv.slice(2));

function checkIfValidPortnumber(num) {
  const regexExp = /^((6553[0-5])|(655[0-2][0-9])|(65[0-4][0-9]{2})|(6[0-4][0-9]{3})|([1-5][0-9]{4})|([0-5]{0,5})|([0-9]{1,4}))$/gi;
  return regexExp.test(num);
}

// const portCandidate = argv.p || 3000;
var selectedPort = argv.p || 3000;

if (selectedPort != 3000) {
  if (checkIfValidPortnumber(selectedPort)){
    // selectedPort = portCandidate;
  } else {
    selectedPort = 3000;
  }
}

const restrictedMode = argv.r || false;


var serverOptions = {};
serverOptions["MinKNOWRunDirectory"] = "";
serverOptions["MARTiSampleDirectory"] = [];
serverOptions["TaxonomyDirectory"] = "";
serverOptions["MaxSimultaneousAnalyses"] = 10;
var numAnalyses = 0;

// var serverOptionsPath = "";
// if(fsExtra.existsSync("./marti_server_options.txt")) {
//   serverOptionsPath = "./marti_server_options.txt";
// } else if(fsExtra.existsSync(homedir + "/marti_server_options.txt")) {
//   serverOptionsPath = homedir + "/marti_server_options.txt";
// } else if (fsExtra.existsSync(homedir + "/.marti_server_options.txt")) {
//   serverOptionsPath = homedir + "/.marti_server_options.txt";
// } else {
//   console.log("Warning: Could not find marti_server_options.txt.");
//   console.log("You must have the file marti_server_options.txt in your home directory to start new analyses.");
// }

// try {
//   const MARTiServerOptions = fsExtra.readFileSync(serverOptionsPath, 'UTF-8');
//   const lines = MARTiServerOptions.split(/\r?\n/);
//   lines.forEach((line) => {
//       if(line.charAt(0) != '#') {
//         const fields = line.split("\t");
//         if(fields[0] == "MinKNOWRunDirectory") {
//           serverOptions["MinKNOWRunDirectory"] = fields[1];
//         } else if(fields[0] == "MARTiSampleDirectory") {
//           const dirs = fields[1].split(":");
//           for (const dir of dirs) {
//             var finalDir;
//             if (dir.endsWith('/')){
//               finalDir = dir.slice(0, -1);
//             } else {
//               finalDir = dir;
//             };
//             serverOptions["MARTiSampleDirectory"].push(finalDir);
//           }
//
//         } else if (fields[0] == "BlastDatabaseDirectory") {
//           serverOptions["BlastDatabaseDirectory"] = fields[1];
//         } else if (fields[0] == "TaxonomyDirectory") {
//           serverOptions["TaxonomyDirectory"] = fields[1];
//         } else if (fields[0] == "MaxSimultaneousAnalyses") {
//           serverOptions["MaxSimultaneousAnalyses"] = parseInt(fields[1]);
//         }
//       }
//   });
//   if( serverOptions["MinKNOWRunDirectory"] == "" ||
//       serverOptions["MARTiSampleDirectory"].length < 1 ||
//       serverOptions["BlastDatabaseDirectory"] == "" ||
//       serverOptions["TaxonomyDirectory"] == "") {
//     console.log("Warning: Could not find all fields in " + serverOptionsPath + ".");
//     console.log("Please check this file and restart to start new analyses.");
//   }
// } catch (err) {
//
// }


var engineOptionsPath = "";
if (argv.options) {
  if(fsExtra.existsSync(argv.options)) {
    engineOptionsPath = argv.options;
  }
} else if(fsExtra.existsSync("./marti_engine_options.txt")) {
    engineOptionsPath = "./marti_engine_options.txt";
  } else if(fsExtra.existsSync(homedir + "/marti_engine_options.txt")) {
    engineOptionsPath = homedir + "/marti_engine_options.txt";
  } else if (fsExtra.existsSync(homedir + "/.marti_engine_options.txt")) {
    engineOptionsPath = homedir + "/.marti_engine_options.txt";
  } else {
    console.log("Warning: Could not find marti_engine_options.txt");
    // console.log("You must have the file marti_engine_options.txt in your home directory to start new analyses.");
  }


engineOptionsObject = {processes:[]};

try {
  const martiEngineOptions = fsExtra.readFileSync(engineOptionsPath, 'UTF-8');
  const lines = martiEngineOptions.split(/\r?\n/);
  var newProcess = false;
  var processFound = false;
  var currentProcess = {text:""};
  lines.forEach((line) => {
      if(line.charAt(0) != '#') {

        if (newProcess == true) {
          if (line == "") {
            newProcess = false;
            engineOptionsObject.processes.push(currentProcess);
            currentProcess = {text:""};
          } else {
            var key = line.split(":")[0].trim();
            var value;
            if (key == "UseToClassify") {
              currentProcess.text += "    " + key + "\n";
              value = "true";
            } else {
              value = line.split(":")[1].trim();
              currentProcess.text += "    " + key + ":" + value + "\n";
            }
            currentProcess[key] = value;
          }
        } else if (line.search("BlastProcess") != -1) {
          newProcess = true;
          processFound = true;
        } else {
          const fields = line.split(":");
          if(fields[0] == "MinKNOWRunDirectory") {
            serverOptions["MinKNOWRunDirectory"] = fields[1];
          } else if(fields[0] == "MARTiSampleDirectory") {
            const dirs = fields[1].split(";");
            for (const dir of dirs) {
              var finalDir;
              if (dir.endsWith('/')){
                finalDir = dir.slice(0, -1);
              } else {
                finalDir = dir;
              };
              serverOptions["MARTiSampleDirectory"].push(finalDir);
            }

          } else if (fields[0] == "TaxonomyDir") {
            serverOptions["TaxonomyDirectory"] = fields[1];
          } else if (fields[0] == "MaxSimultaneousAnalyses") {
            serverOptions["MaxSimultaneousAnalyses"] = parseInt(fields[1]);
          }
        };
      }
  });
  if( serverOptions["MinKNOWRunDirectory"] == "" ||
      serverOptions["MARTiSampleDirectory"].length < 1 ||
      serverOptions["TaxonomyDirectory"] == "") {
    console.log("Warning: Could not find all fields in " + serverOptionsPath + ".");
    console.log("Please check this file and restart to start new analyses.");
  }
  if (newProcess == true) {
    newProcess = false;
    engineOptionsObject.processes.push(currentProcess);
    currentProcess = {text:""};
  }


if(processFound == false) {
    console.log("Warning: Could not find any processes in " + engineOptionsPath);
}
} catch (err) {

}



function getSubDirectories(path) {
  return fsExtra.readdirSync(path).filter(function (file) {
    return fsExtra.statSync(path+'/'+file).isDirectory();
  });
}

function scanMinKNOWRunDirectory() {
  var minKNOWSampleNames = [];
  if(serverOptions["MinKNOWRunDirectory"] != "") {
    try {
      const MinKNOWRunDirectory = serverOptions["MinKNOWRunDirectory"];
      var sampleDirs = [];
      var list = getSubDirectories(MinKNOWRunDirectory);
      list.forEach(function(dir) {
        var newList = getSubDirectories(MinKNOWRunDirectory + "/" + dir);
        if(newList.includes("fastq_pass") || newList.includes("pass")) {
          minKNOWSampleNames.push(MinKNOWRunDirectory + "/" + dir);
        }
        else if(newList.length == 1) {
          const sampleDir = newList[0];
          newList = getSubDirectories(MinKNOWRunDirectory + "/" + dir + "/" + sampleDir);
          if(newList.length == 1) {
            const uid_dir = newList[0];
            newList = getSubDirectories(MinKNOWRunDirectory + "/" + dir + "/" + sampleDir + "/" + uid_dir);
            if(newList.includes("fastq_pass")) {
              minKNOWSampleNames.push(MinKNOWRunDirectory + "/" + dir + "/" + sampleDir + "/" + uid_dir);
            }
          }
        }
      });
    } catch (err) {
      console.error(err);
    }
  }
  serverOptions["minKNOWSampleNames"] = minKNOWSampleNames;
}

function makeConfigFileString(form_object) {
  var configFileString = "";
  configFileString += "RawDataDir:" + form_object["rawDataDir"] + "\n";
  configFileString += "SampleName:" + form_object["martiName"] + "\n";
  configFileString += "SampleDir:" + form_object["outputDir"] + "/" + form_object["martiName"] + "\n";
  if(form_object.hasOwnProperty('processBarcodeCheck') && form_object["processBarcodeCheck"] == "on") {
    if(Array.isArray(form_object["barcodeCheck"])) {
      configFileString += "ProcessBarcodes:";
      for(var i = 0; i < form_object["barcodeCheck"].length; i++) {
        configFileString += form_object["barcodeCheck"][i] + ",";
      }
      configFileString = configFileString.slice(0, -1) + "\n";
      for(var i = 0; i < form_object["barcodeName"].length; i++) {
        configFileString += "BarcodeId" + parseInt(form_object["barcodeCheck"][i]).toString() + ":" + form_object["barcodeName"][i] + "\n";
      }
    } else {
      configFileString += "ProcessBarcodes:" + form_object["barcodeCheck"] + "\n";
      configFileString += "BarcodeId" + parseInt(form_object["barcodeCheck"]).toString() + ":" + form_object["barcodeName"] + "\n";
    }
  }
  configFileString += "Scheduler:" + "local" + "\n";
  configFileString += "LocalSchedulerMaxJobs:" + form_object["maxJobs"] + "\n";
  configFileString += "InactivityTimeout:" + form_object["inactivityTimeout"] + "\n";
  configFileString += "StopProcessingAfter:" + form_object["stopProcessingAfter"] + "\n";
  configFileString += "TaxonomyDir:" + serverOptions["TaxonomyDirectory"] + "\n";
  configFileString += "LCAMaxHits:" + form_object["maxHits"] + "\n";
  configFileString += "LCAScorePercent:" + form_object["scorePercent"] + "\n";
  configFileString += "LCAMinIdentity:" + form_object["minimumIdentity"] + "\n";
  configFileString += "LCAMinQueryCoverage:" + form_object["minQueryCoverage"] + "\n";
  configFileString += "LCAMinCombinedScore:" + form_object["minCombinedScore"] + "\n";
  configFileString += "ReadsPerBlast:" + form_object["readsPerChunk"] + "\n";
  configFileString += "ReadFilterMinQ:" + form_object["readFilterMinQ"] + "\n";
  configFileString += "ReadFilterMinLength:" + form_object["minimumReadLength"] + "\n";

  if(form_object.hasOwnProperty("analysisName")) {
        if(Array.isArray(form_object["analysisName"])) {
          for (var [i, process] of form_object["analysisName"].entries()) {
              configFileString += process;
              configFileString += "\n";
            }
        } else {
              configFileString += form_object["analysisName"];
              configFileString += "\n";
        }
  }

  // if(form_object.hasOwnProperty("processName")) {
  //   if(Array.isArray(form_object["processName"])) {
  //   for(var i = 0; i < form_object["processName"].length; i++) {
  //       configFileString += "BlastProcess\n";
  //       configFileString += "\tName:" + form_object["processName"][i] + "\n";
  //       configFileString += "\tProgram:" + form_object["blastProgram"][i] + "\n";
  //       configFileString += "\tDatabase:" + form_object["databaseDir"][i] + "/" + form_object["blastDatabase"][i] + "\n";
  //       if(form_object["taxaFilter"][i].length > 0) {
  //         configFileString += "\tTaxaFilter:" + form_object["taxaFilter"][i] + "\n";
  //       }
  //       configFileString += "\tMaxE:" + form_object["maxE"][i] + "\n";
  //       configFileString += "\tMaxTargetSeqs:" + form_object["maxTargetSeqs"][i] + "\n";
  //       configFileString += "\tBlastThreads:" + form_object["blastThreads"][i] + "\n";
  //       if(form_object.hasOwnProperty("useToClassify") && form_object["useToClassify"][i] == "on") {
  //         configFileString += "\tUseToClassify\n";
  //       }
  //     }
  //   } else {
  //     configFileString += "BlastProcess\n";
  //     configFileString += "\tName:" + form_object["processName"] + "\n";
  //     configFileString += "\tProgram:" + form_object["blastProgram"] + "\n";
  //     configFileString += "\tDatabase:" + form_object["databaseDir"] + "/" + form_object["blastDatabase"] + "\n";
  //     if(form_object["taxaFilter"].length > 0 ) {
  //       configFileString += "\tTaxaFilter:" + form_object["taxaFilter"] + "\n";
  //     }
  //     configFileString += "\tMaxE:" + form_object["maxE"] + "\n";
  //     configFileString += "\tMaxTargetSeqs:" + form_object["maxTargetSeqs"] + "\n";
  //     configFileString += "\tBlastThreads:" + form_object["blastThreads"] + "\n";
  //     if(form_object.hasOwnProperty("useToClassify") && form_object["useToClassify"] == "on") {
  //       configFileString += "\tUseToClassify\n";
  //     }
  //   }
  // }
  return configFileString;
}


var Observer = require('./services/observer');

app.use(express.static(__dirname + '/public'));

app.use(express.urlencoded({ extended: true }));

var observer = new Observer();


observer.on('id-file-added', meta => {
  metaDataUpdateId(meta);
});

observer.on('id-file-removed', meta => {

  var runId = meta.runId;

  if(sampleMetaDict[runId]){

    for (const [name, newName] of Object.entries(sampleNamesDict[runId])) {
      if (sampleMetaDict[runId][name]) {
        sampleMetaDict[runId][name]["sample"]["id"] = sampleMetaDict[runId][name]["sample"]["originalId"];
      }
    }

    io.sockets.emit('meta-update-available', {
      runId: runId,
      sampleId: ""
    });
  }

    delete sampleNamesDict[runId];

});

function metaDataUpdate(meta) {

var data = meta.content;
var sampleId = meta.id;
var runId = meta.runId;

if (!data.sample.hasOwnProperty("runId")) {
  data.sample.runId = runId;
}


  if (sampleMetaDict[runId]) {
    sampleMetaDict[runId][sampleId] = data;
  } else {
    sampleMetaDict[runId] = {};
    sampleMetaDict[runId][sampleId] = data;
  }
  if(sampleNamesDict[runId]){
    updateMetaDataSampleNames(runId);
  }

  io.sockets.emit('meta-update-available', {
    runId: runId,
    sampleId: sampleId
  });
}

function updateMetaDataSampleNames(runId) {

    for (const [name, newName] of Object.entries(sampleNamesDict[runId])) {
      if (sampleMetaDict[runId][name]) {
        sampleMetaDict[runId][name]["sample"]["id"] = newName;
        sampleMetaDict[runId][name]["sample"]["originalId"] = name;

      } else {

      }
    }


}

var runIdsToUpdate = [];

function metaDataUpdateId(meta) {

  var data = meta.content;
  var runId = meta.runId;

  sampleNamesDict[runId] = data;

  if(sampleMetaDict[runId]){
  updateMetaDataSampleNames(runId);

  io.sockets.emit('meta-update-available', {
    runId: runId,
    sampleId: ""
  });
  }

}

observer.on('meta-file-added', meta => {
  metaDataUpdate(meta);
});

observer.on('meta-file-updated', meta => {
  metaDataUpdate(meta);
});

observer.on('meta-file-removed', meta => {

  var sampleId = meta.id;
  var runId = meta.runName;

  if (sampleMetaDict.hasOwnProperty(runId) && sampleMetaDict[runId].hasOwnProperty(sampleId)) {
    delete sampleMetaDict[runId][sampleId];
  }

  if (sampleTreeDict.hasOwnProperty(runId) && sampleTreeDict[runId].hasOwnProperty(sampleId)) {
    delete sampleTreeDict[runId][sampleId];
  }

  if (sampleAccumulationDict.hasOwnProperty(runId) && sampleAccumulationDict[runId].hasOwnProperty(sampleId)) {
    delete sampleAccumulationDict[runId][sampleId];
  }

  if (sampleAmrDict.hasOwnProperty(runId) && sampleAmrDict[runId].hasOwnProperty(sampleId)) {
    delete sampleAmrDict[runId][sampleId];
  }

  for (const [uuid, entryData] of Object.entries(clientData)) {
    if (sampleId == entryData.selectedDashboardSample.name && runId == entryData.selectedDashboardSample.runId) {
      entryData.selectedDashboardSample.name = "";
      entryData.selectedDashboardSample.runId = "";
    };

    var findSampleInCompare = entryData.selectedCompareSamples.findIndex(e => e.name == sampleId && e.runId == runId);

    if (findSampleInCompare != -1) {
      entryData.selectedCompareSamples.splice(findSampleInCompare,1);
    };

  };

  io.sockets.emit('sample-removed', {
    runId: runId,
    sampleId: sampleId
  });


});

observer.on('tree-file-added', tree => {


  if (sampleTreeDict[tree.runName]) {
    if (sampleTreeDict[tree.runName][tree.id]) {
      sampleTreeDict[tree.runName][tree.id][tree.lca]=tree.content;
    } else {
      sampleTreeDict[tree.runName][tree.id] = {};
      sampleTreeDict[tree.runName][tree.id][tree.lca]=tree.content;
    }
  } else {
    sampleTreeDict[tree.runName] = {};
    sampleTreeDict[tree.runName][tree.id] = {};
    sampleTreeDict[tree.runName][tree.id][tree.lca]=tree.content;
  }


});

observer.on('tree-file-updated', tree => {
  sampleTreeDict[tree.runName][tree.id][tree.lca]=tree.content;

  for (const [id, data] of Object.entries(clientData)) {
    if ((tree.id == data.selectedDashboardSample.name && tree.runName == data.selectedDashboardSample.runId) || (data.selectedCompareSamples.filter(e => e.name == tree.id && e.runId == tree.runName).length > 0)) {
      io.to(id).emit('tree-update-available');
      console.log(`[${new Date().toLocaleString()}][${id}] Tree update notification sent`);
    };
  };

});



observer.on('accumulation-file-added', data => {

  var accumulationData = data.content;
  var id = data.id;
  var runId = data.runName;
  var lca = data.lca;

    if (sampleAccumulationDict[runId]) {
      if (sampleAccumulationDict[runId][id]) {
        sampleAccumulationDict[runId][id][lca]=accumulationData;
      } else {
        sampleAccumulationDict[runId][id] = {};
        sampleAccumulationDict[runId][id][lca]=accumulationData;
      }
    } else {
      sampleAccumulationDict[runId] = {};
      sampleAccumulationDict[runId][id] = {};
      sampleAccumulationDict[runId][id][lca]=accumulationData;
    }

    for (const [uuid, entryData] of Object.entries(clientData)) {
      if ((id == entryData.selectedDashboardSample.name && runId == entryData.selectedDashboardSample.runId) || entryData.selectedCompareSamples.filter(e => e.name == id && e.runId == runId).length > 0) {
        io.to(uuid).emit('accumulation-update-available');
        console.log(`[${new Date().toLocaleString()}][${uuid}] Accumulation update notification sent`);
      };
    };

});

observer.on('accumulation-file-updated', data => {

  var accumulationData = data.content;
  var id = data.id;
  var runId = data.runName;
  var lca = data.lca;

sampleAccumulationDict[runId][id][lca]=accumulationData;

  for (const [uuid, entryData] of Object.entries(clientData)) {
    if ((id == entryData.selectedDashboardSample.name && runId == entryData.selectedDashboardSample.runId) || entryData.selectedCompareSamples.filter(e => e.name == id && e.runId == runId).length > 0) {
      io.to(uuid).emit('accumulation-update-available');
      console.log(`[${new Date().toLocaleString()}][${uuid}] Accumulation update notification sent`);
    };
  };

});


observer.on('amr-file-added', data => {

  var amrData = data.content;
  var id = data.id;
  var runId = data.runName;

  if (sampleAmrDict[runId]) {
    sampleAmrDict[runId][id] = amrData;
  } else {
    sampleAmrDict[runId] = {};
    sampleAmrDict[runId][id] = amrData;
  }


  for (const [uuid, entryData] of Object.entries(clientData)) {
    if (id == entryData.selectedDashboardSample.name && runId == entryData.selectedDashboardSample.runId) {
      io.to(uuid).emit('amr-update-available');
      console.log(`[${new Date().toLocaleString()}][${uuid}] Amr update notification sent`);
    };
  };


});

observer.on('amr-file-updated', data => {

  var amrData = data.content;
  var id = data.id;
  var runId = data.runName;

  sampleAmrDict[runId][id] = amrData;

  for (const [uuid, entryData] of Object.entries(clientData)) {
    if ((id == entryData.selectedDashboardSample.name && runId == entryData.selectedDashboardSample.runId) || entryData.selectedCompareSamples.filter(e => e.name == id && e.runId == runId).length > 0) {
      io.to(uuid).emit('amr-update-available');
      console.log(`[${new Date().toLocaleString()}][${uuid}] Amr update notification sent`);
    };
  };

});

var martiDirs = [];
var projectsEnabled = false;
var projectsDatabase = {};

if(serverOptions["MARTiSampleDirectory"].length > 0){
  for (var dir of serverOptions["MARTiSampleDirectory"]) {
    if (!dir.endsWith("/")) {
      dir = dir + "/";
    };

    try {
      const projectsObj = fsExtra.readJsonSync(dir + "projects.json");
      for (const [project,runs] of Object.entries(projectsObj)){
        projectsDatabase[project] = runs;
      }
      projectsEnabled = true;
    } catch (err) {
    }

    observer.watchFolder(dir);
  }

} else {
  console.log("MARTiSampleDirectory not specified.");
}


var sampleMetaDict = {};
var sampleNamesDict = {};
var sampleTreeDict = {};
var sampleAccumulationDict = {};
var sampleAmrDict = {};

var database = [];


var clientData = {};


app.get('/', function(req, res){
  res.sendFile(__dirname + '/indexNode.html');
});

http.listen(selectedPort, '0.0.0.0', function(){
  console.log(`[${new Date().toLocaleString()}] listening on port ${selectedPort}`);
});



app.get('/:runId/:sampleId/:lca/csv', function (req, res) {
  var run = req.params.runId;
  var name = req.params.sampleId;
  var lca = req.params.lca;
  var dir = sampleMetaDict[run][name].sample.dir;
  const file = dir + '/'+ run + '/marti/' + name + '/assignments_ms' + lca + '.csv';
  res.download(file);
})

app.get('/project/:project', function (req, res) {
  res.sendFile(__dirname + '/indexNode.html');
})




app.post('/new',(req, res) => {
  if(numAnalyses < serverOptions["MaxSimultaneousAnalyses"] ) {
    // check output dir exists, create if not
    const outputDir = req.body["outputDir"] + "/" + req.body["martiName"];
    if (!fsExtra.existsSync(outputDir)) {
      fsExtra.mkdirSync(outputDir);
    }

    // write the config file
    const configFileString = makeConfigFileString(req.body);
    fsExtra.writeFile(outputDir + "/config.txt", configFileString, err => {
      if (err) {
        console.error(err)
      }
    })

    // start MARTi
    var logStream = fsExtra.createWriteStream(outputDir + "/output.txt", {flags: 'a'});
    var spawn = require('child_process').spawn,
      marti_process = spawn('marti', ['-config', outputDir + '/config.txt']);
      numAnalyses += 1;
      marti_process.stdout.pipe(logStream);
      marti_process.stderr.pipe(logStream);

      marti_process.on('close', function (code) {
        numAnalyses -= 1;
        console.log('child process exited with code ' + code);
      });
    } else {
    console.log("Max number of analyses reached. Could not start new analysis.");
  }

  res.end();
});

function sendHeartbeat(){
    io.sockets.emit('hb_ping', { beat : 1 });
    setTimeout(sendHeartbeat, 8000);
}


var clientCount;

io.on('connect', function(socket){

  clientCount = socket.client.conn.server.clientsCount;
  console.log(`[${new Date().toLocaleString()}] Connection added - users connected: ${clientCount}`);
  io.sockets.emit('current-client-count', clientCount);

  socket.on('hb_pong', function(data){
  });

  socket.on('register-request', function(data){ // a client requests registration
    var id;

    if (data.uuid == null) {
      id = uuidv4();
    } else {
      id = data.uuid;
    }

    if(!clientData.hasOwnProperty(id)) {
      clientData[id] = {
        selectedDashboardSample: {
          name: data.currentDashboardSampleName,
          runId: data.currentDashboardSampleRun
        },
        selectedCompareSamples: data.compareSampleObjectArray,
        project: data.clientProject
      };
      console.log(`[${new Date().toLocaleString()}][${id}] New client ID`);
    } else {
      console.log(`[${new Date().toLocaleString()}][${id}] Already exists in object`);
    }

      socket.join(id);
      // io.to(id).emit('register-response', id);
      io.to(id).emit('register-response', {
        id: id,
        mode: restrictedMode
      });

      console.log(`[${new Date().toLocaleString()}][${id}] Client registered`);
      sendHeartbeat();


  });


  socket.on('meta-request', request => {
    var id = request.clientId;
    if (!projectsEnabled){
    io.to(id).emit('meta-response', sampleMetaDict);
    console.log(`[${new Date().toLocaleString()}][${id}] Metadata sent`);
  } else {
    if (clientData[id]["project"]){
      var project = clientData[id]["project"];
      if(projectsDatabase.hasOwnProperty(project)){
        var customMetaDict = {};
        for (const [run,samples] of Object.entries(sampleMetaDict)){
          if(projectsDatabase[project].includes(run)){
            customMetaDict[run] = samples;
          }
        }
        io.to(id).emit('meta-response', customMetaDict);
        console.log(`[${new Date().toLocaleString()}][${id}] Metadata sent`);
      }

    }


  }
  });

  socket.on('update-sample-name-request', request => {
    var newId = request.newId;
    var originalId = request.originalId;
    var sampleId = request.pathName;
    var runId = request.pathRun;
    var dir;
    var idFileContent={};
      if (sampleMetaDict[runId]) {
        dir = sampleMetaDict[runId][sampleId]["sample"]["dir"]
        var idFilePath = dir + "/" + runId + "/ids.json";

        if (fsExtra.existsSync(idFilePath)) {
          idFileContent = fsExtra.readFileSync(idFilePath);
          idFileContent = JSON.parse(idFileContent);
        }

        idFileContent[originalId] = newId;

        fsExtra.writeJson(idFilePath, idFileContent, (err) => {
          if (err) throw err;

        });

      }

  })

  socket.on('default-server-options-request', request => {
    var id = request.clientId;
    scanMinKNOWRunDirectory();
    serverOptions.processes = engineOptionsObject.processes;
    io.to(id).emit('default-server-options-response', serverOptions);
    console.log(`[${new Date().toLocaleString()}][${id}] Default server options sent`);
  });

  socket.on('selected-dashboard-sample', sample => {
        var id = sample.clientId;
        var sampleId = sample.name;
        var runId = sample.runId;
        clientData[id].selectedDashboardSample = {
          name: sampleId,
          runId: runId
        };
        io.to(id).emit('current-dashboard-sample-response', clientData[id].selectedDashboardSample);
        // console.log(`[${new Date().toLocaleString()}][${id}] Dashboard sample selected: ${runId} - ${sampleId}`);
  });

  socket.on('selected-compare-samples', samples => {
      var id = samples.clientId;
      var samples = samples.data;
      clientData[id].selectedCompareSamples = samples;
      io.to(id).emit('current-compare-samples-response', clientData[id].selectedCompareSamples);
      var sampleNames = [];
      for (const sample of samples) {
        sampleNames.push(sample.name);
      }
      // console.log(`[${new Date().toLocaleString()}][${id}] Compare samples selected: ${sampleNames}`);
  });

  socket.on('current-dashboard-sample-request', request => {
      var id = request.clientId;
      io.to(id).emit('current-dashboard-sample-response', clientData[id].selectedDashboardSample);
  });

  socket.on('current-compare-samples-request', request => {
      var id = request.clientId;
      io.to(id).emit('current-compare-samples-response', clientData[id].selectedCompareSamples);
  });

  socket.on('dashboard-tree-request', request => {
    var id = request.clientId;
    var lca = request.lca;
    var selectedDashboardSample = clientData[id].selectedDashboardSample.name;
    var selectedDashboardRun = clientData[id].selectedDashboardSample.runId;

    if (sampleTreeDict.hasOwnProperty(selectedDashboardRun)) {
          if (sampleTreeDict[selectedDashboardRun].hasOwnProperty(selectedDashboardSample)) {
            io.to(id).emit('dashboard-tree-response', {
              id: selectedDashboardSample,
              run: selectedDashboardRun,
              treeData: sampleTreeDict[selectedDashboardRun][selectedDashboardSample][lca],
              treeData2: sampleTreeDict[selectedDashboardRun][selectedDashboardSample][lca]
              });
            console.log(`[${new Date().toLocaleString()}][${id}] Dashboard tree data sent at lca: ${lca}`);
          }
        };

    });

  socket.on('dashboard-meta-request', request => {
    var id = request.clientId;
    var selectedDashboardSampleName = clientData[id].selectedDashboardSample.name;
    var selectedDashboardSampleRunId = clientData[id].selectedDashboardSample.runId;
    io.to(id).emit('dashboard-meta-response', sampleMetaDict[selectedDashboardSampleRunId][selectedDashboardSampleName]);
    console.log(`[${new Date().toLocaleString()}][${id}] Dashboard meta data sent`);
    });

socket.on('compare-tree-request', request => {
  var id = request.clientId;
  var lca = request.lca;
  var selectedCompareSamples = clientData[id].selectedCompareSamples;
  compareTreeData = [];
  for (var sample of selectedCompareSamples){
    if (sampleTreeDict.hasOwnProperty(sample.runId)) {
      if (sampleTreeDict[sample.runId].hasOwnProperty(sample.name)) {
        compareTreeData.push({
          id: sample.name,
          runId: sample.runId,
          tree: sampleTreeDict[sample.runId][sample.name][lca]
        });
        };
      };
    };
  io.to(id).emit('compare-tree-response', compareTreeData);
  console.log(`[${new Date().toLocaleString()}][${id}] Compare tree data sent at lca: ${lca}`);
  });

  socket.on('compare-accumulation-request', request => {
    var id = request.clientId;
    var rank = request.rank;
    var lca = request.lca;

    var compareAccumulationData = [];
    var selectedCompareSamples = clientData[id].selectedCompareSamples;

    for (var sample of selectedCompareSamples){
      if (sampleAccumulationDict.hasOwnProperty(sample.runId)) {
        if (sampleAccumulationDict[sample.runId].hasOwnProperty(sample.name)) {
          compareAccumulationData.push({
            id: sample.name,
            runId: sample.runId,
            data: sampleAccumulationDict[sample.runId][sample.name][lca][rank]
          });
        }
      }
    };


    if (compareAccumulationData.length != 0) {
      io.to(id).emit('compare-accumulation-response', compareAccumulationData);
      console.log(`[${new Date().toLocaleString()}][${id}] Compare accumulation data sent`);
    };

    });

    socket.on('compare-amr-request', request => {
      var id = request.clientId;
      var selectedCompareSamples = clientData[id].selectedCompareSamples;
      var compareAmrData = [];
      for (var sample of selectedCompareSamples){
        if (sampleAmrDict.hasOwnProperty(sample.runId)) {
          if (sampleAmrDict[sample.runId].hasOwnProperty(sample.name)) {
            compareAmrData.push({
              id: sample.name,
              runId: sample.runId,
              data: sampleAmrDict[sample.runId][sample.name]
            });
            };
          };
        };
      if (compareAmrData.length != 0) {
      io.to(id).emit('compare-amr-response', compareAmrData);
      console.log(`[${new Date().toLocaleString()}][${id}] Compare amr data sent`);
      };
      });

    socket.on('dashboard-accumulationChart-request', request => {
      var id = request.clientId;
      var rank = request.rank;
      var lca = request.lca;


      var selectedDashboardSample = clientData[id].selectedDashboardSample.name;
      var selectedDashboardRun = clientData[id].selectedDashboardSample.runId;

      if (sampleAccumulationDict.hasOwnProperty(selectedDashboardRun)) {
            if (sampleAccumulationDict[selectedDashboardRun].hasOwnProperty(selectedDashboardSample)) {
              if (sampleAccumulationDict[selectedDashboardRun][selectedDashboardSample].hasOwnProperty(lca)) {
                io.to(id).emit('dashboard-accumulationChart-response', [{
                  id: selectedDashboardSample,
                  runId: selectedDashboardRun,
                  data: sampleAccumulationDict[selectedDashboardRun][selectedDashboardSample][lca][rank]
                }]);
                console.log(`[${new Date().toLocaleString()}][${id}] Dashboard accumulation data sent`);
              }
            }
          };

      });

      socket.on('dashboard-dashboardAmrTable-request', request => {
        var uuid = request.clientId;

        var selectedDashboardSample = clientData[uuid].selectedDashboardSample.name;
        var selectedDashboardRun = clientData[uuid].selectedDashboardSample.runId;

        if (sampleAmrDict.hasOwnProperty(selectedDashboardRun)) {
              if (sampleAmrDict[selectedDashboardRun].hasOwnProperty(selectedDashboardSample)) {
                io.to(uuid).emit('dashboard-dashboardAmrTable-response', sampleAmrDict[selectedDashboardRun][selectedDashboardSample]);
                console.log(`[${new Date().toLocaleString()}][${uuid}] Dashboard amr data sent`);
              }
            };

        });



        socket.on('disconnect', () => {
          clientCount = socket.client.conn.server.clientsCount;
          console.log(`[${new Date().toLocaleString()}] Connection removed - users connected: ${clientCount}`);
          io.sockets.emit('current-client-count', clientCount);
        });

        socket.on('disconnecting', () => {
          const rooms = Object.keys(socket.rooms);
          var property = rooms[1];
          console.log(`[${new Date().toLocaleString()}] Client disconnecting: ${property}`);

        });



});
